class Entity:
    def total_number_of_named_entities(
        SE: object
        ) -> float:
        try:
            return SE.total_number_of_named_entities_ 
        except AttributeError:
            doc_total = 0
            for ent in SE.doc.ents:
                doc_total += 1
            SE.total_number_of_named_entities_ = doc_total
            return SE.total_number_of_named_entities_



# -*- coding: UTF-8 -*-
"""
Software: LFTK - Linguistic Features Toolkit
Script: extractor.py
License: CC-BY-SA 4.0
Original Author: Bruce W. Lee (이웅성) @brucewlee
Affiliation 1: LXPER Inc. - Seoul, South Korea
Affiliation 2: University of Pennsylvania - PA, USA
"""
import os
from typing import Union, Dict, List

import spacy
import pandas as pd

from lftk.utils import get_feature_map, get_pandas_row, convert_ndjson_to_pd
from lftk.derivation_collector import DerivationCollector


CURRENT_PATH = os.path.abspath(os.path.dirname(__file__))
FEATURE_MAP_PATH = os.path.join(CURRENT_PATH, 'resources/feature_map.json')
AOA_KUP_PATH = os.path.join(CURRENT_PATH, 'resources/AoA_ratings_kup.csv')
AOA_BRY_PATH = os.path.join(CURRENT_PATH, 'resources/AoA_ratings_bry.csv')
SUBTLEX_US_PATH = os.path.join(CURRENT_PATH, 'resources/subtlex_us.csv')


class Extractor:
    def __init__(
        self, 
        docs: Union[spacy.tokens.doc.Doc, List[spacy.tokens.doc.Doc]]
        ) -> None:
        """
        Starter class for users
        input :
        - self
        - docs: single or multiple spacy doc object
        saves :
        - self.doc: spacy doc object
        - self.feature_map: dictionary of dictionary that contains "key" as the parent key and the others ("name", "type") as children keys
        """
        # Type adjustment
        if type(docs) is not list: docs = [docs]
        self.docs = docs
        self.feature_map = get_feature_map(path = FEATURE_MAP_PATH)
        self.customize()

    def customize(
        self,
        stop_words: bool = True,
        punctuations: bool = True,
        round_decimal: int = 3,
        ) -> None:
        """
        Save options
        input :
        - self
        - options
        saves :
        - self.options
        """
        self.options = {
            'stop_words': stop_words,
            'punctuations': punctuations,
            'round_decimal': 3
        }

    def extract(
        self, 
        features: Union[str, list] = "*"
        ) -> Union[
            Dict[str, Union[float,int]], 
            List[Dict[str, Union[float,int]]]
            ]:
        """
        Calculate linguistic feature(s) from all given spaCy docs
        input :
        - self
        - features: features to be extracted
        returns :
        - result: extracted linguistic feature(s)
        """
        # Type adjustment
        if type(features) is not list: features = [features]
        # Iterate user input
        all_results = []
        for doc in self.docs:
            extractor = SingleExtractor(doc, self.feature_map, self.options)
            result = extractor.run(features)
            all_results.append(result)
        # Type adjustment
        if len(all_results) == 1:
            all_results = all_results[0]
        return all_results


class SingleExtractor:
    def __init__(
        self, 
        doc: spacy.tokens.doc.Doc,
        feature_map: Dict[str, Dict[str, str]],
        options: Dict[str, Union[str, bool]]
        ) -> None:
        """
        Initialize pipeline
        input :
        - self
        - doc: single text inside spacy doc object
        saves :
        - self.doc: spacy doc object
        """
        self.doc = doc
        self.feature_map = feature_map
        self.options = options
        self.paths = {
            'FEATURE_MAP_PATH' : FEATURE_MAP_PATH,
            'AOA_KUP_PATH' : AOA_KUP_PATH,
            'AOA_BRY_PATH' : AOA_BRY_PATH,
            'SUBTLEX_US_PATH' : SUBTLEX_US_PATH,
        }

    def run(
        self, 
        features: list,
        ) -> Dict[str, Union[float,int]]:
        """
        Calculate linguistic feature(s) from spaCy doc
        input :
        - self
        - features: features to be extracted
        returns :
        - result: extracted linguistic feature(s)
        """
        # Iterate through given features
        result = {}
        if features != ["*"]:
            for feature_key in features:
                # Get feature name from feature key
                feature_name = self.feature_map[feature_key]['name']
                # Get feature function from feature name
                feature_fn = getattr(DerivationCollector, feature_name)
                # Run feature function in DerivationCollector object
                result[f'{feature_key}'] = round(feature_fn(self), self.options['round_decimal'])
        else:
            # Extract all features
            for feature_key in self.feature_map:
                # Get feature name from feature key
                feature_name = self.feature_map[feature_key]['name']
                # Get feature function from feature name
                feature_fn = getattr(DerivationCollector, feature_name)
                # Run feature function in DerivationCollector object
                result[f'{feature_key}'] = round(feature_fn(self), self.options['round_decimal'])
        return result

def search_features(
    domain: str = "*",
    family: str = "*",
    pandas: bool = False
    ) -> Union[List[dict], pd.core.frame.DataFrame]:
    """
    Return available linguistic features
    input :
    - domain: specify which domain
    - family: specify which family
    returns :
    - result: searched linguistic feature(s)
    """
    feature_df = convert_ndjson_to_pd(path = FEATURE_MAP_PATH)
    if domain != "*":
        feature_df = get_pandas_row(feature_df, "domain", domain, True)
    if family != "*":
        feature_df = get_pandas_row(feature_df, "family", family, True)
    if pandas == False:
        result = feature_df.to_dict('records')
    else:
        result = feature_df
    return result


from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class AvgEntity(FoundationCollector):
    def average_number_of_named_entities_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_named_entities_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_named_entities_ = \
                FoundationCollector.total_number_of_named_entities(SE)
            SE.average_number_of_named_entities_per_word_ = \
                safe_division(
                    total_number_of_named_entities_,
                    total_number_of_words_
                )
            return SE.average_number_of_named_entities_per_word_
    

    def average_number_of_named_entities_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_named_entities_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_named_entities_ = \
                FoundationCollector.total_number_of_named_entities(SE)
            SE.average_number_of_named_entities_per_sentence_ = \
                safe_division(
                    total_number_of_named_entities_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_named_entities_per_sentence_


class PartOfSpeech:
    def total_number_of_adjectives(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_adjectives_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_adjectives_ = pos_list.count("ADJ")
            return SE.total_number_of_adjectives_

    def total_number_of_adpositions(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_adpositions_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_adpositions_ = pos_list.count("ADP")
            return SE.total_number_of_adpositions_
    
    def total_number_of_adverbs(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_adverbs_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_adverbs_ = pos_list.count("ADV")
            return SE.total_number_of_adverbs_
    
    def total_number_of_auxiliaries(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_auxiliaries_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_auxiliaries_ = pos_list.count("AUX")
            return SE.total_number_of_auxiliaries_

    def total_number_of_conjunctions(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_conjunctions_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_conjunctions_ = pos_list.count("CONJ")
            return SE.total_number_of_conjunctions_

    def total_number_of_coordinating_conjunctions(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_coordinating_conjunctions_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_coordinating_conjunctions_ = pos_list.count("CCONJ")
            return SE.total_number_of_coordinating_conjunctions_

    def total_number_of_determiners(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_determiners_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_determiners_ = pos_list.count("DET")
            return SE.total_number_of_determiners_

    def total_number_of_interjections(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_interjections_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_interjections_ = pos_list.count("INTJ")
            return SE.total_number_of_interjections_
    
    def total_number_of_nouns(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_nouns_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_nouns_ = pos_list.count("NOUN")
            return SE.total_number_of_nouns_

    def total_number_of_numerals(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_numerals_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_numerals_ = pos_list.count("NUM")
            return SE.total_number_of_numerals_

    def total_number_of_particles(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_particles_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_particles_ = pos_list.count("PART")
            return SE.total_number_of_particles_

    def total_number_of_pronouns(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_pronouns_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_pronouns_ = pos_list.count("PRON")
            return SE.total_number_of_pronouns_

    def total_number_of_proper_nouns(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_proper_nouns_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_proper_nouns_ = pos_list.count("PROPN")
            return SE.total_number_of_proper_nouns_

    def total_number_of_punctuations(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_punctuations_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_punctuations_ = pos_list.count("PUNCT")
            return SE.total_number_of_punctuations_

    def total_number_of_subordinating_conjunctions(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_subordinating_conjunctions_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_subordinating_conjunctions_ = pos_list.count("SCONJ")
            return SE.total_number_of_subordinating_conjunctions_

    def total_number_of_symbols(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_symbols_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_symbols_ = pos_list.count("SYM")
            return SE.total_number_of_symbols_
    
    def total_number_of_verbs(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_verbs_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_verbs_ = pos_list.count("VERB")
            return SE.total_number_of_verbs_

    def total_number_of_spaces(
        SE: object
        ) -> float:
        try: 
            return SE.total_number_of_verbs_
        except AttributeError:
            pos_list = [token.pos_ for token in SE.doc]
            SE.total_number_of_verbs_ = pos_list.count("SPACE")
            return SE.total_number_of_verbs_


import math

import spacy
from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class ReadFormula(FoundationCollector):
    def flesch_kincaid_reading_ease(
        SE: object,
        ) -> float:
        """
        flesch kincaid reading ease formula
        100.00–90.00	5th grade	        Very easy to read.
        90.0–80.0	    6th grade	        Easy to read.
        80.0–70.0	    7th grade	        Fairly easy to read.
        70.0–60.0	    8th & 9th grade	    Plain English. 
        60.0–50.0	    10th to 12th grade	Fairly difficult to read.
        50.0–30.0	    College	            Difficult to read.
        30.0–10.0	    College graduate	Very difficult to read. 
        10.0–0.0	    Professional	    Extremely difficult to read. 
        input :
        - SE: Single Extractor object
        output:
        - SE.flesch_kincaid_reading_ease_
        """
        try:
            return SE.flesch_kincaid_reading_ease_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_syllables_ = \
                FoundationCollector.total_number_of_syllables(SE)
            average_number_of_words_per_sentence_ = \
                safe_division(
                    total_number_of_words_,
                    total_number_of_sentences_
                )
            average_number_of_syllables_per_word_ = \
                safe_division(
                    total_number_of_syllables_,
                    total_number_of_words_
                )
            # Calculate readability
            SE.flesch_kincaid_reading_ease_ = \
                206.835 - 1.015 * average_number_of_words_per_sentence_ - 84.6 * average_number_of_syllables_per_word_
            # Revert back to user options
            SE.options = original_options
            return SE.flesch_kincaid_reading_ease_
    
    def flesch_kincaid_grade_level(
        SE: object,
        ) -> float:
        """
        flesch kincaid grade level
        output number corresponds to US grade level
        input :
        - SE: Single Extractor object
        output:
        - SE.flesch_kincaid_grade_level_
        """
        try:
            return SE.flesch_kincaid_grade_level_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_syllables_ = \
                FoundationCollector.total_number_of_syllables(SE)
            average_number_of_words_per_sentence_ = \
                safe_division(
                    total_number_of_words_,
                    total_number_of_sentences_
                )
            average_number_of_syllables_per_word_ = \
                safe_division(
                    total_number_of_syllables_,
                    total_number_of_words_
                )
            # Calculate readability
            SE.flesch_kincaid_grade_level_ = \
                0.39 * average_number_of_words_per_sentence_ + 11.8 * average_number_of_syllables_per_word_ - 15.59
            # Revert back to user options
            SE.options = original_options
            return SE.flesch_kincaid_grade_level_
    
    def gunning_fog_index(
        SE: object,
        ) -> float:
        """
        gunning fog index
        17	College graduate
        16	College senior
        15	College junior
        14	College sophomore
        13	College freshman
        12	High school senior
        11	High school junior
        10	High school sophomore
        9	High school freshman
        8	Eighth grade
        7	Seventh grade
        6	Sixth grade
        input :
        - SE: Single Extractor object
        output:
        - SE.gunning_fog_index_
        """
        try:
            return SE.gunning_fog_index_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_words_more_than_two_syllables_ = \
                FoundationCollector.total_number_of_words_more_than_two_syllables(SE)
            average_number_of_words_per_sentence_ = \
                safe_division(
                    total_number_of_words_,
                    total_number_of_sentences_
                )
            average_number_of_words_more_than_two_syllables_per_word_ = \
                safe_division(
                    total_number_of_words_more_than_two_syllables_,
                    total_number_of_words_
                )
            # Calculate readability
            SE.gunning_fog_index_ = \
                0.4 * (average_number_of_words_per_sentence_ + 100 * average_number_of_words_more_than_two_syllables_per_word_)
            # Revert back to user options
            SE.options = original_options
            return SE.gunning_fog_index_
    
    def smog_index(
        SE: object,
        ) -> float:
        """
        simple measure of gobbledygook
        output number corresponds to US grade level
        input :
        - SE: Single Extractor object
        output:
        - SE.smog_index_
        """
        try:
            return SE.smog_index_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_words_more_than_three_syllables_ = \
                FoundationCollector.total_number_of_words_more_than_three_syllables(SE)
            # Calculate readability
            SE.smog_index_ = \
                1.0430 * math.sqrt(total_number_of_words_more_than_three_syllables_ * safe_division(30, total_number_of_sentences_))
            # Revert back to user options
            SE.options = original_options
            return SE.smog_index_

    def coleman_liau_index(
        SE: object,
        ) -> float:
        """
        coleman liau index
        output number corresponds to US grade level
        input :
        - SE: Single Extractor object
        output:
        - SE.coleman_liau_index_
        """
        try:
            return SE.coleman_liau_index_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_characters_ = \
                FoundationCollector.total_number_of_characters(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            average_number_of_letters_per_100words_ = \
                safe_division(
                    total_number_of_characters_,
                    total_number_of_words_
                ) * 100
            average_number_of_sentences_per_100words_ = \
                safe_division(
                    total_number_of_sentences_,
                    total_number_of_words_
                ) * 100
            # Calculate readability
            SE.coleman_liau_index_ = \
                0.0588 * average_number_of_letters_per_100words_ - 0.296 * average_number_of_sentences_per_100words_ - 15.8
            # Revert back to user options
            SE.options = original_options
            return SE.coleman_liau_index_
    
    def automated_readability_index(
        SE: object,
        ) -> float:
        """
        automated readability index
        output number corresponds to US grade level
        input :
        - SE: Single Extractor object
        output:
        - SE.automated_readability_index_
        """
        try:
            return SE.automated_readability_index_
        except AttributeError:
            # Override user options
            original_options = SE.options
            SE.options['stop_words'] = True
            SE.options['punctuations'] = False
            # Calculate necessary foundation features
            total_number_of_characters_ = \
                FoundationCollector.total_number_of_characters(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            average_number_of_letters_per_word_ = \
                safe_division(
                    total_number_of_characters_,
                    total_number_of_words_
                )
            average_number_of_words_per_sentence_ = \
                safe_division(
                    total_number_of_words_,
                    total_number_of_sentences_
                )
            # Calculate readability
            SE.automated_readability_index_ = \
                4.71 * average_number_of_letters_per_word_ + 0.5 * average_number_of_words_per_sentence_ - 21.43
            # Revert back to user options
            SE.options = original_options
            return SE.automated_readability_index_


import pandas as pd

from lftk.utils import get_pandas_row, safe_division

class WordDiff:
    def total_kuperman_age_of_acquistion_of_words(
        SE: object
        ) -> float:
        try:
            return SE.total_kuperman_age_of_acquistion_of_words_ 
        except AttributeError:
            # Load databse
            database = pd.read_csv(SE.paths['AOA_KUP_PATH'])
            # Iterate doc
            doc_total = 0
            for token in SE.doc:
                row = get_pandas_row(database, 'Word', token.lemma_)
                try:
                    doc_total += float(row[0]['Rating.Mean'])
                except IndexError:
                    pass
            SE.total_kuperman_age_of_acquistion_of_words_ = doc_total
            return SE.total_kuperman_age_of_acquistion_of_words_


    def total_brysbaert_age_of_acquistion_of_words(
        SE: object
        ) -> float:
        try:
            return SE.total_brysbaert_age_of_acquistion_of_words_ 
        except AttributeError:
            database = pd.read_csv(SE.paths['AOA_BRY_PATH'])
            doc_total = 0
            for token in SE.doc:
                rows = get_pandas_row(database, 'WORD', token.lemma_)
                aoa_by_token_meaning = []
                for row in rows:
                    try:
                        aoa_by_token_meaning.append(row['AoAtestbased'])
                    except IndexError:
                        pass
                doc_total += safe_division(
                    sum(aoa_by_token_meaning),len(aoa_by_token_meaning)
                    )
            SE.total_brysbaert_age_of_acquistion_of_words_ = doc_total
            return SE.total_brysbaert_age_of_acquistion_of_words_

    
    def total_subtlex_us_zipf_of_words(
        SE: object
        ) -> float:
        try:
            return SE.total_subtlex_us_zipf_of_words_ 
        except AttributeError:
            database = pd.read_csv(SE.paths['SUBTLEX_US_PATH'])
            doc_total = 0
            for token in SE.doc:
                row = get_pandas_row(database, 'Word', token.text.lower())
                try:
                    doc_total += float(row[0]['Zipf-value'])
                except IndexError:
                    pass
            SE.total_subtlex_us_zipf_of_words_ = doc_total
            return SE.total_subtlex_us_zipf_of_words_



from lftk.derivation.avgwordsent import AvgWordSent
from lftk.derivation.avgworddiff import AvgWordDiff
from lftk.derivation.avgentity   import AvgEntity
from lftk.derivation.typetokenratio import TypeTokenRatio
from lftk.derivation.avgpartofspeech import AvgPartOfSpeech
from lftk.derivation.readformula import ReadFormula

class DerivationCollector(AvgWordSent, AvgWordDiff, AvgEntity, TypeTokenRatio, AvgPartOfSpeech, ReadFormula):
    pass 


import math
from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class TypeTokenRatio(FoundationCollector):
    def simple_type_token_ratio(
        SE: object,
        ) -> float:
        try:
            return SE.simple_type_token_ratio_
        except AttributeError:
            total_number_of_unique_words_ = \
                FoundationCollector.total_number_of_unique_words(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.simple_type_token_ratio_ = \
                safe_division(
                    total_number_of_unique_words_,
                    total_number_of_words_
                )
            return SE.simple_type_token_ratio_
    
    def root_type_token_ratio(
        SE: object,
        ) -> float:
        try:
            return SE.root_type_token_ratio_
        except AttributeError:
            total_number_of_unique_words_ = \
                FoundationCollector.total_number_of_unique_words(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.root_type_token_ratio_ = \
                safe_division(
                    total_number_of_unique_words_,
                    math.sqrt(total_number_of_words_)
                )
            return SE.root_type_token_ratio_
    
    def corrected_type_token_ratio(
        SE: object,
        ) -> float:
        try:
            return SE.corrected_type_token_ratio_
        except AttributeError:
            total_number_of_unique_words_ = \
                FoundationCollector.total_number_of_unique_words(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.corrected_type_token_ratio_ = \
                safe_division(
                    total_number_of_unique_words_,
                    math.sqrt(2*total_number_of_words_)
                )
            return SE.corrected_type_token_ratio_

    def bilogarithmic_type_token_ratio(
        SE: object,
        ) -> float:
        try:
            return SE.bilogarithmic_type_token_ratio_
        except AttributeError:
            total_number_of_unique_words_ = \
                FoundationCollector.total_number_of_unique_words(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.bilogarithmic_type_token_ratio_ = \
                safe_division(
                    math.log(total_number_of_unique_words_),
                    math.log(total_number_of_words_)
                )
            return SE.bilogarithmic_type_token_ratio_
    
    def uber_type_token_ratio(
        SE: object,
        ) -> float:
        try:
            return SE.uber_type_token_ratio_
        except AttributeError:
            total_number_of_unique_words_ = \
                FoundationCollector.total_number_of_unique_words(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.uber_type_token_ratio_ = \
                safe_division(
                    math.log(total_number_of_unique_words_)**2,
                    math.log(
                        safe_division(
                            total_number_of_words_, 
                            total_number_of_unique_words_
                        )
                    )
                )
            return SE.uber_type_token_ratio_

    def simple_type_token_ratio_no_lemma(
        SE: object,
        ) -> float:
        try:
            return SE.simple_type_token_ratio_no_lemma_
        except AttributeError:
            total_number_of_unique_words_no_lemma_ = \
                FoundationCollector.total_number_of_unique_words_no_lemma(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.simple_type_token_ratio_no_lemma_ = \
                safe_division(
                    total_number_of_unique_words_no_lemma_,
                    total_number_of_words_
                )
            return SE.simple_type_token_ratio_no_lemma_
    
    def root_type_token_ratio_no_lemma(
        SE: object,
        ) -> float:
        try:
            return SE.root_type_token_ratio_no_lemma_
        except AttributeError:
            total_number_of_unique_words_no_lemma_ = \
                FoundationCollector.total_number_of_unique_words_no_lemma(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.root_type_token_ratio_no_lemma_ = \
                safe_division(
                    total_number_of_unique_words_no_lemma_,
                    math.sqrt(total_number_of_words_)
                )
            return SE.root_type_token_ratio_no_lemma_
    
    def corrected_type_token_ratio_no_lemma(
        SE: object,
        ) -> float:
        try:
            return SE.corrected_type_token_ratio_no_lemma_
        except AttributeError:
            total_number_of_unique_words_no_lemma_ = \
                FoundationCollector.total_number_of_unique_words_no_lemma(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.corrected_type_token_ratio_no_lemma_ = \
                safe_division(
                    total_number_of_unique_words_no_lemma_,
                    math.sqrt(2*total_number_of_words_)
                )
            return SE.corrected_type_token_ratio_no_lemma_

    def bilogarithmic_type_token_ratio_no_lemma(
        SE: object,
        ) -> float:
        try:
            return SE.bilogarithmic_type_token_ratio_no_lemma_
        except AttributeError:
            total_number_of_unique_words_no_lemma_ = \
                FoundationCollector.total_number_of_unique_words_no_lemma(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.bilogarithmic_type_token_ratio_no_lemma_ = \
                safe_division(
                    math.log(total_number_of_unique_words_no_lemma_),
                    math.log(total_number_of_words_)
                )
            return SE.bilogarithmic_type_token_ratio_no_lemma_
    
    def uber_type_token_ratio_no_lemma(
        SE: object,
        ) -> float:
        try:
            return SE.uber_type_token_ratio_no_lemma_
        except AttributeError:
            total_number_of_unique_words_no_lemma_ = \
                FoundationCollector.total_number_of_unique_words_no_lemma(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.uber_type_token_ratio_no_lemma_ = \
                safe_division(
                    math.log(total_number_of_unique_words_no_lemma_)**2,
                    math.log(
                        safe_division(
                            total_number_of_words_, 
                            total_number_of_unique_words_no_lemma_
                        )
                    )
                )
            return SE.uber_type_token_ratio_no_lemma_



from lftk.foundation.wordsent import WordSent
from lftk.foundation.worddiff import WordDiff
from lftk.foundation.entity   import Entity
from lftk.foundation.partofspeech import PartOfSpeech

class FoundationCollector(WordSent, WordDiff, Entity, PartOfSpeech):
    pass


from typing import Union, List

import ndjson
import pandas as pd

def get_feature_map(path):
    feature_df = convert_ndjson_to_pd(path).set_index('key')
    feature_map = feature_df.to_dict('index')
    return feature_map

def convert_ndjson_to_pd(path: str):
    with open(path, encoding='utf-8') as f:	
        data = ndjson.load(f)
        data = pd.DataFrame(data)
    return data

def get_pandas_row(
    dataframe: pd.core.frame.DataFrame, 
    column: str, 
    to_search: str,
    return_pandas: bool = False
    ) -> Union[List[dict], pd.core.frame.DataFrame]:
    mask = dataframe[column] == to_search
    if return_pandas == False:
        rows = dataframe[mask].to_dict('records')
    else:
        rows = dataframe[mask]
    return rows

def safe_division(n, d):
    return n / d if d else 0

if __name__ == "__main__":
    df = convert_ndjson_to_pd('resources/feature_map.json')
    df.to_csv('feature_map.csv', index = False)


import spacy
from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class AvgWordDiff(FoundationCollector):
    def average_kuperman_age_of_acquistion_of_words_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_kuperman_age_of_acquistion_of_words_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_kuperman_age_of_acquistion_of_words_ = \
                FoundationCollector.total_kuperman_age_of_acquistion_of_words(SE)
            SE.average_kuperman_age_of_acquistion_of_words_per_sentence_ = \
                safe_division(
                    total_kuperman_age_of_acquistion_of_words_,
                    total_number_of_sentences_
                    )
            return SE.average_kuperman_age_of_acquistion_of_words_per_sentence_


    def average_kuperman_age_of_acquistion_of_words_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_kuperman_age_of_acquistion_of_words_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_kuperman_age_of_acquistion_of_words_ = \
                FoundationCollector.total_kuperman_age_of_acquistion_of_words(SE)
            SE.average_kuperman_age_of_acquistion_of_words_per_word_ = \
                safe_division(
                    total_kuperman_age_of_acquistion_of_words_,
                    total_number_of_words_
                    )
            return SE.average_kuperman_age_of_acquistion_of_words_per_word_
    

    def average_brysbaert_age_of_acquistion_of_words_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_brysbaert_age_of_acquistion_of_words_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_brysbaert_age_of_acquistion_of_words_ = \
                FoundationCollector.total_brysbaert_age_of_acquistion_of_words(SE)
            SE.average_brysbaert_age_of_acquistion_of_words_per_word_ = \
                safe_division(
                    total_brysbaert_age_of_acquistion_of_words_,
                    total_number_of_words_
                )
            return SE.average_brysbaert_age_of_acquistion_of_words_per_word_
    

    def average_brysbaert_age_of_acquistion_of_words_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_brysbaert_age_of_acquistion_of_words_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_brysbaert_age_of_acquistion_of_words_ = \
                FoundationCollector.total_brysbaert_age_of_acquistion_of_words(SE)
            SE.average_brysbaert_age_of_acquistion_of_words_per_sentence_ = \
                safe_division(
                    total_brysbaert_age_of_acquistion_of_words_,
                    total_number_of_sentences_
                )
            return SE.average_brysbaert_age_of_acquistion_of_words_per_sentence_

    
    def average_subtlex_us_zipf_of_words_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_subtlex_us_zipf_of_words_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_subtlex_us_zipf_of_words_ = \
                FoundationCollector.total_subtlex_us_zipf_of_words(SE)
            SE.average_subtlex_us_zipf_of_words_per_word_ = \
                safe_division(
                    total_subtlex_us_zipf_of_words_,
                    total_number_of_words_
                )
            return SE.average_subtlex_us_zipf_of_words_per_word_
    

    def average_subtlex_us_zipf_of_words_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_subtlex_us_zipf_of_words_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_subtlex_us_zipf_of_words_ = \
                FoundationCollector.total_subtlex_us_zipf_of_words(SE)
            SE.subtlex_us_zipf_of_words_per_sentence_ = \
                safe_division(
                    total_subtlex_us_zipf_of_words_,
                    total_number_of_sentences_
                )
            return SE.subtlex_us_zipf_of_words_per_sentence_


from distutils.core import setup
from setuptools import find_packages
# python setup.py sdist
# python -m twine upload dist/*
setup(
    name = 'lftk',  
    version='1.0.3',
    license='CC BY-NC 4.0',   
    description = 'Comprehensive Multilingual Linguistic Features Extraction in Python',
    author = 'Bruce W. Lee',
    author_email = 'brucelws@seas.upenn.edu', 
    packages=find_packages(),
    url='https://github.com/brucewlee/lftk',
    keywords='linguistic feature',
    install_requires=[
          'pandas',
          'ndjson',
          'spacy'
      ],
    include_package_data=True,
    package_data={'': ['resources/*']}
)


import spacy
from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class AvgWordSent(FoundationCollector):
    def average_number_of_words_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_words_per_sentence_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            SE.average_number_of_words_per_sentence_ = \
                safe_division(
                    total_number_of_words_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_words_per_sentence_
    
    def average_number_of_characters_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_characters_per_sentence_
        except AttributeError:
            total_number_of_characters_ = \
                FoundationCollector.total_number_of_characters(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            SE.average_number_of_characters_per_sentence_ = \
                safe_division(
                    total_number_of_characters_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_characters_per_sentence_

    def average_number_of_characters_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_characters_per_word_
        except AttributeError:
            total_number_of_characters_ = \
                FoundationCollector.total_number_of_characters(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.average_number_of_characters_per_word_ = \
                safe_division(
                    total_number_of_characters_,
                    total_number_of_words_
                )
            return SE.average_number_of_characters_per_word_

    def average_number_of_syllables_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_syllables_per_sentence_
        except AttributeError:
            total_number_of_syllables_ = \
                FoundationCollector.total_number_of_syllables(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            SE.average_number_of_syllables_per_sentence_ = \
                safe_division(
                    total_number_of_syllables_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_syllables_per_sentence_

    def average_number_of_syllables_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_syllables_per_word_
        except AttributeError:
            total_number_of_syllables_ = \
                FoundationCollector.total_number_of_syllables(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.average_number_of_syllables_per_word_ = \
                safe_division(
                    total_number_of_syllables_,
                    total_number_of_words_
                )
            return SE.average_number_of_syllables_per_word_
    
    def average_number_of_stop_words_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_stop_words_per_sentence_
        except AttributeError:
            total_number_of_stop_words_ = \
                FoundationCollector.total_number_of_stop_words(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            SE.average_number_of_stop_words_per_sentence_ = \
                safe_division(
                    total_number_of_stop_words_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_stop_words_per_sentence_

    def average_number_of_stop_words_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_stop_words_per_word_
        except AttributeError:
            total_number_of_stop_words_ = \
                FoundationCollector.total_number_of_stop_words(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.average_number_of_stop_words_per_word_ = \
                safe_division(
                    total_number_of_stop_words_,
                    total_number_of_words_
                )
            return SE.average_number_of_stop_words_per_word_

    def average_number_of_punctuations_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_punctuations_per_sentence_
        except AttributeError:
            total_number_of_punctuations_ = \
                FoundationCollector.total_number_of_punctuations(SE)
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            SE.average_number_of_punctuations_per_sentence_ = \
                safe_division(
                    total_number_of_punctuations_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_punctuations_per_sentence_

    def average_number_of_punctuations_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_punctuations_per_word_
        except AttributeError:
            total_number_of_punctuations_ = \
                FoundationCollector.total_number_of_punctuations(SE)
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            SE.average_number_of_punctuations_per_word_ = \
                safe_division(
                    total_number_of_punctuations_,
                    total_number_of_words_
                )
            return SE.average_number_of_punctuations_per_word_


from lftk.derivation.foundation_collector import FoundationCollector
from lftk.utils import safe_division

class AvgPartOfSpeech(FoundationCollector):
    def average_number_of_adjectives_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_adjectives_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_adjectives_ = \
                FoundationCollector.total_number_of_adjectives(SE)
            SE.average_number_of_adjectives_per_word_ = \
                safe_division(
                    total_number_of_adjectives_,
                    total_number_of_words_
                )
            return SE.average_number_of_adjectives_per_word_
    
    def average_number_of_adjectives_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_adjectives_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_adjectives_ = \
                FoundationCollector.total_number_of_adjectives(SE)
            SE.average_number_of_adjectives_per_sentence_ = \
                safe_division(
                    total_number_of_adjectives_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_adjectives_per_sentence_

    def average_number_of_adpositions_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_adpositions_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_adpositions_ = \
                FoundationCollector.total_number_of_adpositions(SE)
            SE.average_number_of_adpositions_per_word_ = \
                safe_division(
                    total_number_of_adpositions_,
                    total_number_of_words_
                )
            return SE.average_number_of_adpositions_per_word_
    
    def average_number_of_adpositions_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_adpositions_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_adpositions_ = \
                FoundationCollector.total_number_of_adpositions(SE)
            SE.average_number_of_adpositions_per_sentence_ = \
                safe_division(
                    total_number_of_adpositions_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_adpositions_per_sentence_
    
    def average_number_of_adverbs_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_adverbs_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_adverbs_ = \
                FoundationCollector.total_number_of_adverbs(SE)
            SE.average_number_of_adverbs_per_word_ = \
                safe_division(
                    total_number_of_adverbs_,
                    total_number_of_words_
                )
            return SE.average_number_of_adverbs_per_word_
    
    def average_number_of_adverbs_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_adverbs_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_adverbs_ = \
                FoundationCollector.total_number_of_adverbs(SE)
            SE.average_number_of_adverbs_per_sentence_ = \
                safe_division(
                    total_number_of_adverbs_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_adverbs_per_sentence_
    
    def average_number_of_auxiliaries_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_auxiliaries_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_auxiliaries_ = \
                FoundationCollector.total_number_of_auxiliaries(SE)
            SE.average_number_of_auxiliaries_per_word_ = \
                safe_division(
                    total_number_of_auxiliaries_,
                    total_number_of_words_
                )
            return SE.average_number_of_auxiliaries_per_word_
    
    def average_number_of_auxiliaries_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_auxiliaries_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_auxiliaries_ = \
                FoundationCollector.total_number_of_auxiliaries(SE)
            SE.average_number_of_auxiliaries_per_sentence_ = \
                safe_division(
                    total_number_of_auxiliaries_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_auxiliaries_per_sentence_
    
    def average_number_of_conjunctions_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_conjunctions_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_conjunctions_ = \
                FoundationCollector.total_number_of_conjunctions(SE)
            SE.average_number_of_conjunctions_per_word_ = \
                safe_division(
                    total_number_of_conjunctions_,
                    total_number_of_words_
                )
            return SE.average_number_of_conjunctions_per_word_
    
    def average_number_of_conjunctions_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_conjunctions_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_conjunctions_ = \
                FoundationCollector.total_number_of_conjunctions(SE)
            SE.average_number_of_conjunctions_per_sentence_ = \
                safe_division(
                    total_number_of_conjunctions_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_conjunctions_per_sentence_

    def average_number_of_coordinating_conjunctions_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_coordinating_conjunctions_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_coordinating_conjunctions_ = \
                FoundationCollector.total_number_of_coordinating_conjunctions(SE)
            SE.average_number_of_coordinating_conjunctions_per_word_ = \
                safe_division(
                    total_number_of_coordinating_conjunctions_,
                    total_number_of_words_
                )
            return SE.average_number_of_coordinating_conjunctions_per_word_
    
    def average_number_of_coordinating_conjunctions_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_coordinating_conjunctions_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_coordinating_conjunctions_ = \
                FoundationCollector.total_number_of_coordinating_conjunctions(SE)
            SE.average_number_of_coordinating_conjunctions_per_sentence_ = \
                safe_division(
                    total_number_of_coordinating_conjunctions_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_coordinating_conjunctions_per_sentence_

    def average_number_of_determiners_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_determiners_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_determiners_ = \
                FoundationCollector.total_number_of_determiners(SE)
            SE.average_number_of_determiners_per_word_ = \
                safe_division(
                    total_number_of_determiners_,
                    total_number_of_words_
                )
            return SE.average_number_of_determiners_per_word_
    
    def average_number_of_determiners_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_determiners_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_determiners_ = \
                FoundationCollector.total_number_of_determiners(SE)
            SE.average_number_of_determiners_per_sentence_ = \
                safe_division(
                    total_number_of_determiners_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_determiners_per_sentence_

    def average_number_of_interjections_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_interjections_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_interjections_ = \
                FoundationCollector.total_number_of_interjections(SE)
            SE.average_number_of_interjections_per_word_ = \
                safe_division(
                    total_number_of_interjections_,
                    total_number_of_words_
                )
            return SE.average_number_of_interjections_per_word_
    
    def average_number_of_interjections_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_interjections_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_interjections_ = \
                FoundationCollector.total_number_of_interjections(SE)
            SE.average_number_of_interjections_per_sentence_ = \
                safe_division(
                    total_number_of_interjections_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_interjections_per_sentence_

    def average_number_of_nouns_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_nouns_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_nouns_ = \
                FoundationCollector.total_number_of_nouns(SE)
            SE.average_number_of_nouns_per_word_ = \
                safe_division(
                    total_number_of_nouns_,
                    total_number_of_words_
                )
            return SE.average_number_of_nouns_per_word_
    
    def average_number_of_nouns_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_nouns_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_nouns_ = \
                FoundationCollector.total_number_of_nouns(SE)
            SE.average_number_of_nouns_per_sentence_ = \
                safe_division(
                    total_number_of_nouns_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_nouns_per_sentence_

    def average_number_of_numerals_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_numerals_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_numerals_ = \
                FoundationCollector.total_number_of_numerals(SE)
            SE.average_number_of_numerals_per_word_ = \
                safe_division(
                    total_number_of_numerals_,
                    total_number_of_words_
                )
            return SE.average_number_of_numerals_per_word_
    
    def average_number_of_numerals_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_numerals_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_numerals_ = \
                FoundationCollector.total_number_of_numerals(SE)
            SE.average_number_of_numerals_per_sentence_ = \
                safe_division(
                    total_number_of_numerals_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_numerals_per_sentence_

    def average_number_of_particles_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_particles_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_particles_ = \
                FoundationCollector.total_number_of_particles(SE)
            SE.average_number_of_particles_per_word_ = \
                safe_division(
                    total_number_of_particles_,
                    total_number_of_words_
                )
            return SE.average_number_of_particles_per_word_
    
    def average_number_of_particles_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_particles_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_particles_ = \
                FoundationCollector.total_number_of_particles(SE)
            SE.average_number_of_particles_per_sentence_ = \
                safe_division(
                    total_number_of_particles_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_particles_per_sentence_

    def average_number_of_pronouns_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_pronouns_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_pronouns_ = \
                FoundationCollector.total_number_of_pronouns(SE)
            SE.average_number_of_pronouns_per_word_ = \
                safe_division(
                    total_number_of_pronouns_,
                    total_number_of_words_
                )
            return SE.average_number_of_pronouns_per_word_
    
    def average_number_of_pronouns_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_pronouns_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_pronouns_ = \
                FoundationCollector.total_number_of_pronouns(SE)
            SE.average_number_of_pronouns_per_sentence_ = \
                safe_division(
                    total_number_of_pronouns_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_pronouns_per_sentence_

    def average_number_of_proper_nouns_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_proper_nouns_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_proper_nouns_ = \
                FoundationCollector.total_number_of_proper_nouns(SE)
            SE.average_number_of_proper_nouns_per_word_ = \
                safe_division(
                    total_number_of_proper_nouns_,
                    total_number_of_words_
                )
            return SE.average_number_of_proper_nouns_per_word_
    
    def average_number_of_proper_nouns_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_proper_nouns_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_proper_nouns_ = \
                FoundationCollector.total_number_of_proper_nouns(SE)
            SE.average_number_of_proper_nouns_per_sentence_ = \
                safe_division(
                    total_number_of_proper_nouns_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_proper_nouns_per_sentence_

    def average_number_of_punctuations_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_punctuations_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_punctuations_ = \
                FoundationCollector.total_number_of_punctuations(SE)
            SE.average_number_of_punctuations_per_word_ = \
                safe_division(
                    total_number_of_punctuations_,
                    total_number_of_words_
                )
            return SE.average_number_of_punctuations_per_word_
    
    def average_number_of_punctuations_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_punctuations_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_punctuations_ = \
                FoundationCollector.total_number_of_punctuations(SE)
            SE.average_number_of_punctuations_per_sentence_ = \
                safe_division(
                    total_number_of_punctuations_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_punctuations_per_sentence_

    def average_number_of_subordinating_conjunctions_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_subordinating_conjunctions_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_subordinating_conjunctions_ = \
                FoundationCollector.total_number_of_subordinating_conjunctions(SE)
            SE.average_number_of_subordinating_conjunctions_per_word_ = \
                safe_division(
                    total_number_of_subordinating_conjunctions_,
                    total_number_of_words_
                )
            return SE.average_number_of_subordinating_conjunctions_per_word_
    
    def average_number_of_subordinating_conjunctions_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_subordinating_conjunctions_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_subordinating_conjunctions_ = \
                FoundationCollector.total_number_of_subordinating_conjunctions(SE)
            SE.average_number_of_subordinating_conjunctions_per_sentence_ = \
                safe_division(
                    total_number_of_subordinating_conjunctions_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_subordinating_conjunctions_per_sentence_

    def average_number_of_symbols_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_symbols_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_symbols_ = \
                FoundationCollector.total_number_of_symbols(SE)
            SE.average_number_of_symbols_per_word_ = \
                safe_division(
                    total_number_of_symbols_,
                    total_number_of_words_
                )
            return SE.average_number_of_symbols_per_word_
    
    def average_number_of_symbols_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_symbols_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_symbols_ = \
                FoundationCollector.total_number_of_symbols(SE)
            SE.average_number_of_symbols_per_sentence_ = \
                safe_division(
                    total_number_of_symbols_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_symbols_per_sentence_

    def average_number_of_verbs_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_verbs_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_verbs_ = \
                FoundationCollector.total_number_of_verbs(SE)
            SE.average_number_of_verbs_per_word_ = \
                safe_division(
                    total_number_of_verbs_,
                    total_number_of_words_
                )
            return SE.average_number_of_verbs_per_word_
    
    def average_number_of_verbs_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_verbs_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_verbs_ = \
                FoundationCollector.total_number_of_verbs(SE)
            SE.average_number_of_verbs_per_sentence_ = \
                safe_division(
                    total_number_of_verbs_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_verbs_per_sentence_

    def average_number_of_spaces_per_word(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_spaces_per_word_
        except AttributeError:
            total_number_of_words_ = \
                FoundationCollector.total_number_of_words(SE)
            total_number_of_spaces_ = \
                FoundationCollector.total_number_of_spaces(SE)
            SE.average_number_of_spaces_per_word_ = \
                safe_division(
                    total_number_of_spaces_,
                    total_number_of_words_
                )
            return SE.average_number_of_spaces_per_word_
    
    def average_number_of_spaces_per_sentence(
        SE: object,
        ) -> float:
        try:
            return SE.average_number_of_spaces_per_sentence_
        except AttributeError:
            total_number_of_sentences_ = \
                FoundationCollector.total_number_of_sentences(SE)
            total_number_of_spaces_ = \
                FoundationCollector.total_number_of_spaces(SE)
            SE.average_number_of_spaces_per_sentence_ = \
                safe_division(
                    total_number_of_spaces_,
                    total_number_of_sentences_
                )
            return SE.average_number_of_spaces_per_sentence_


class WordSent:
    def total_number_of_words(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_words_ 
        except AttributeError:   
            token_list = [token for token in SE.doc]
            if not SE.options['stop_words']:
                token_list = [token for token in token_list if token.is_stop != True]
            if not SE.options['punctuations']:
                token_list = [token for token in token_list if token.is_punct != True]
            # Calculate and save result
            SE.total_number_of_words_ = len(token_list)
            return SE.total_number_of_words_
    
    def total_number_of_stop_words(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_stop_words_ 
        except AttributeError:   
            token_list = [token for token in SE.doc if token.is_stop == True]
            # Calculate and save result
            SE.total_number_of_stop_words_  = len(token_list)
            return SE.total_number_of_stop_words_ 
    
    def total_number_of_puntuations(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_puntuations_ 
        except AttributeError:   
            token_list = [token for token in SE.doc if token.is_punct == True]
            # Calculate and save result
            SE.total_number_of_puntuations_  = len(token_list)
            return SE.total_number_of_puntuations_ 
    
    def total_number_of_syllables(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_syllables_
        except AttributeError:
            SE.total_number_of_syllables_ = 0
            # always remove punctuations
            token_list = [token for token in SE.doc if token.is_punct != True]
            if not SE.options['stop_words']:
                token_list = [token for token in token_list if token.is_stop != True]
            # start counting syllables per word
            for token in token_list:
                word = token.text.lower()
                vowels = "aeiouy"
                # if word starts with a vowel,
                if word[0] in vowels:
                    SE.total_number_of_syllables_ += 1
                for index in range(1, len(word)):
                    # if a character is vowel and the preceding character is not,
                    if word[index] in vowels and word[index - 1] not in vowels:
                        SE.total_number_of_syllables_ += 1
                # if word ends with e,
                if word.endswith("e"):
                    SE.total_number_of_syllables_ -= 1
                # if count was 0,
                if SE.total_number_of_syllables_ == 0:
                    SE.total_number_of_syllables_ += 1
        return SE.total_number_of_syllables_

    def total_number_of_words_more_than_two_syllables(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_words_more_than_two_syllables_
        except AttributeError:
            SE.total_number_of_words_more_than_two_syllables_ = 0
            # always remove punctuations
            token_list = [token for token in SE.doc if token.is_punct != True]
            if not SE.options['stop_words']:
                token_list = [token for token in token_list if token.is_stop != True]
            # start counting syllables per word
            for token in token_list:
                word = token.text.lower()
                total_number_of_syllables_this_word_ = 0
                vowels = "aeiouy"
                # if word starts with a vowel,
                if word[0] in vowels:
                    total_number_of_syllables_this_word_ += 1
                for index in range(1, len(word)):
                    # if a character is vowel and the preceding character is not,
                    if word[index] in vowels and word[index - 1] not in vowels:
                        total_number_of_syllables_this_word_ += 1
                # if word ends with e,
                if word.endswith("e"):
                    total_number_of_syllables_this_word_ -= 1
                # if count was 0,
                if total_number_of_syllables_this_word_ == 0:
                    total_number_of_syllables_this_word_ += 1
                if total_number_of_syllables_this_word_ >= 3:
                    SE.total_number_of_words_more_than_two_syllables_ += 1
        return SE.total_number_of_words_more_than_two_syllables_

    def total_number_of_words_more_than_three_syllables(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_words_more_than_three_syllables_
        except AttributeError:
            SE.total_number_of_words_more_than_three_syllables_ = 0
            # always remove punctuations
            token_list = [token for token in SE.doc if token.is_punct != True]
            if not SE.options['stop_words']:
                token_list = [token for token in token_list if token.is_stop != True]
            # start counting syllables per word
            for token in token_list:
                word = token.text.lower()
                total_number_of_syllables_this_word_ = 0
                vowels = "aeiouy"
                # if word starts with a vowel,
                if word[0] in vowels:
                    total_number_of_syllables_this_word_ += 1
                for index in range(1, len(word)):
                    # if a character is vowel and the preceding character is not,
                    if word[index] in vowels and word[index - 1] not in vowels:
                        total_number_of_syllables_this_word_ += 1
                # if word ends with e,
                if word.endswith("e"):
                    total_number_of_syllables_this_word_ -= 1
                # if count was 0,
                if total_number_of_syllables_this_word_ == 0:
                    total_number_of_syllables_this_word_ += 1
                if total_number_of_syllables_this_word_ >= 4:
                    SE.total_number_of_words_more_than_three_syllables_ += 1
        return SE.total_number_of_words_more_than_three_syllables_

    def total_number_of_unique_words(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_unique_words_ 
        except AttributeError:   
            token_list = [token for token in SE.doc]
            if not SE.options['stop_words']:
                token_list = [token for token in token_list if token.is_stop != True]
            if not SE.options['punctuations']:
                token_list = [token for token in token_list if token.is_punct != True]
            # Count unique tokens in terms of lemma
            unique_token_list = [token.lemma_ for token in SE.doc]
            unique_token_list = list(set(unique_token_list))
            # Calculate and save result
            SE.total_number_of_unique_words_ = len(unique_token_list)
            return SE.total_number_of_unique_words_
    
    def total_number_of_unique_words_no_lemma(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_unique_words_no_lemma_ 
        except AttributeError:   
            token_list = [token for token in SE.doc]
            if not SE.options['stop_words']:
                token_list = [token for token in token_list if token.is_stop != True]
            if not SE.options['punctuations']:
                token_list = [token for token in token_list if token.is_punct != True]
            # Count unique tokens in terms of lemma
            unique_token_list = [token.lemma_ for token in SE.doc]
            unique_token_list = list(set(unique_token_list))
            # Calculate and save result
            SE.total_number_of_unique_words_no_lemma_ = len(unique_token_list)
            return SE.total_number_of_unique_words_no_lemma_

    def total_number_of_sentences(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_sentences_
        except AttributeError:
            # Calculate and save result
            SE.total_number_of_sentences_ =\
                len(list(SE.doc.sents))
            return SE.total_number_of_sentences_
    
    def total_number_of_characters(
        SE: object,
        ) -> int:
        try:
            return SE.total_number_of_characters_
        except AttributeError:
            token_list = [token for token in SE.doc]
            if not SE.options['stop_words']:
                token_list = [token for token in token_list if token.is_stop != True]
            if not SE.options['punctuations']:
                token_list = [token for token in token_list if token.is_punct != True]
            # Calculate and save result
            SE.total_number_of_characters_ =\
                sum([len(token) for token in token_list])
            return SE.total_number_of_characters_
